#include "prim.h"

#include <iostream.h> 
#include <math.h>
#include <stdlib.h>
#include <cstring>

#define INTSQRT(x)	(INTEGER) ceil(sqrt(x))
#define POSITION	(*d != 0) ?*d :GetFirstPos(*p, field)

//---------------------------------------------------------------------------
Prim::Prim(int argc, char** argv) : lprim(NULL)
{
for(int i = 1; i < argc; i++)
	{
	if(strcmp(argv[i], "-?") == 0) this->Help();

	if(strcmp(argv[i], "-sap") == 0) print.SetSaveP();
	if(strcmp(argv[i], "-shp") == 0) print.SetShowP();
	if(strcmp(argv[i], "-sat") == 0) print.SetSaveT();
	if(strcmp(argv[i], "-sht") == 0) print.SetShowT();
	}
}
//---------------------------------------------------------------------------
void Prim::SearchPrims(INTEGER pStart, INTEGER pStop)
{
start = (pStart > 1) ?pStart: 2;
stop  =  pStop  + 1;

STATE = print.bShow();
PRINT = print.bActive();

prims = 0;
twins = 0;
oldpr = 1;

status.Initialize(start, stop);
status.StartTime();

this->LittlePrims();
this->SetArrays();
this->Sieve();
this->GetPrimNr();

status.StopTime();
status.GetTime();
}
//---------------------------------------------------------------------------
void Prim::SetArrays()
{
if(lprim != NULL) delete lprim;

lprim = new INTEGER[mem.size()];
dis   = new INTEGER[mem.size()];

for(INTEGER i = 0; i < mem.size(); i++)
	{
	lprim[i] = mem[i];
	dis  [i] = 0;
	}

mem.clear();
}
//---------------------------------------------------------------------------
INTEGER Prim::GetFirstPos(INTEGER x, INTEGER y)
{
INTEGER i = (x - y % x) % x;
INTEGER j = (y + i) % 2;

return (j != 0) ?i :i + x;
}
//---------------------------------------------------------------------------
inline void Prim::isPrim(INTEGER n)
{
if(PRINT) print.Prime(n);
prims++;

if(n - oldpr == 2)
	{
	if(PRINT) print.Twin(n);
	twins++;
	}

oldpr = n;
}
//---------------------------------------------------------------------------
void Prim::LittlePrims()
{
INTEGER x = INTSQRT(stop);
INTEGER y = INTSQRT(x);

if(start <= 2 && x >= 2) isPrim(2);
if(start <= 3 && x >= 3) isPrim(3);

mem.push_back(3);
mem.push_back(stop);

ITER p;

for(INTEGER n = 5, s; n <= x; n += 2)
	{
	s = INTSQRT(n);
	p = mem.begin();

	while(*p <= s)
		{
		if(n % *p == 0) break;

		p++;
		}

	if(*p <= s) continue;

	mem.insert(mem.end() - 1, n);

	if(n <= y && y > start) isPrim(n);
	}
}
//---------------------------------------------------------------------------
void Prim::Sieve()
{
INTEGER incr  = INTSQRT(stop);
INTEGER field = start;

INTEGER i;
INTEGER j;
INTEGER *p;
INTEGER *d;
INTEGER SQRT;

bool *flags = new bool[incr];
bool *f	    = flags;
bool again  = true;

for(i = 0; i < incr; i++) *f++ = true;


while(again)
	{
	//--------------------------------------------

	i = field + incr;

	if(i < stop) SQRT = INTSQRT(i);
	else
		{
		SQRT  = INTSQRT(stop);
		incr  = stop - field;
		field = stop - incr;
		again = false;
		}

	//--------------------------------------------

	for(p = lprim, d = dis; *p <= SQRT; p++, d++)
		{
		i = POSITION;
		f = flags + i;

		j = *p + *p;

		while(i < incr)
			{
			if(*f) *f = false;

			f += j;
			i += j;
			}

		*d = i - incr;
		}

	//--------------------------------------------

	i = ~field & 1;
	f =  flags + i;

	while(i < incr)
		{
		if(!*f)	*f = true; else isPrim(field + i);

		f += 2;
		i += 2;
		}

	//--------------------------------------------

	field += incr;

	if(STATE) status.GetStatus(field);
	}
}
//---------------------------------------------------------------------------
void Prim::Help()
{
cout << "Prim - Version 4.0, Copyright (c) Kim Walisch 2002\n\n";

cout << "Parameter:\n\n";

cout << "  -?    zeigt die Hilfe an.\n";
cout << "  -shp  zeigt die Primzahlen auf dem Bildschirm an.\n";
cout << "  -sht  zeigt die Primzahlzwillinge auf dem Bildschirm an.\n";
cout << "  -sap  speichert die Primzahlen in eine Text-Datei.\n";
cout << "  -sat  speichert die Primzahlzwillinge in eine Text-Datei.\n\n";

cout << "mailto: kim_walisch@gmx.net";

exit(1);
}
//---------------------------------------------------------------------------
void Prim::GetPrimNr()
{
cout << "\a\n";
cout << "\nAnzahl der gefundenen Primzahlen:        " << prims;
cout << "\nAnzahl der gefundenen Primzahlzwillinge: " << twins;
}
//---------------------------------------------------------------------------
